import AssemblyKeys._

// a version of old build file using the new SBT syntax
// build definition also includes any *.scala files located in /project folder
// objects and classes are not allowed in this file

lazy val apacheSimplex = Seq(
  libraryDependencies += "org.apache.commons" % "commons-math3" % "3.6.1"
)

lazy val scalaCheck = Seq(
  resolvers += "Sonatype Releases" at "http://oss.sonatype.org/content/repositories/releases",
  libraryDependencies += "org.scalacheck" %% "scalacheck" % "1.13.0" % Test //note that the %% adds the scala version after scalacheck
)

lazy val scalaTest = Seq(
  libraryDependencies += "org.scalatest" %% "scalatest" % "2.2.6" % Test
)

//select sub-directories of base that are not src/test
def srcDirs(base: File) = ( (base / "src") * ("*" -- "test") ) filter { _.isDirectory }

// for scalaxy/streams
// Set environment variable SCALAXY_STREAMS_VERY_VERBOSE=1 to see what is going on
// Better enable optimization manually
lazy val scalaxy = Seq(
  // scalacOptions += "-Xplugin-require:scalaxy-streams",
  scalacOptions += "-Xplugin-disable:scalaxy-streams",
  // scalacOptions in Test ~= (_ filterNot (_ == "-Xplugin-require:scalaxy-streams")),
  // scalacOptions in Test += "-Xplugin-disable:scalaxy-streams",
  // autoCompilerPlugins := true,
  // addCompilerPlugin("com.nativelibs4java" %% "scalaxy-streams" % "0.3.4"),
  // libraryDependencies += "com.nativelibs4java" %% "scalaxy-loops" % "0.3.3" % "provided",
  libraryDependencies += "com.nativelibs4java" %% "scalaxy-streams" % "0.3.4" % "provided"
)

lazy val root = (project in file(".")).
  settings(assemblySettings: _*).
  settings(
    name := "beagle",
    version := "0.9",
    scalaVersion := "2.11.8",
    scalaSource in Compile := baseDirectory.value / "src",                           //SBT default is to have source files in src/main/scala
    unmanagedSourceDirectories in Compile := srcDirs(baseDirectory.value).get,
    unmanagedSources in Compile ++= ((baseDirectory.value / "src") * "*.scala").get, //manually add any *.scala source in src/
    unmanagedSourceDirectories in Test += baseDirectory.value / "src" / "test",      //re-add src/test directory when testing
    mainClass in (Compile, run) := Some("beagle.main"),
    retrieveManaged := true,                                                         //default is false
    libraryDependencies += "org.scala-lang.modules" %% "scala-parser-combinators" % "1.0.4",
    scalacOptions ++= Seq(
      "-unchecked",
      "-feature",
      "-deprecation",
      "-language:implicitConversions",
      "-language:postfixOps",
      "-Dscalac.patmat.analysisBudget=620",
      //"-optimise,
      "-Yclosure-elim",
      "-Yinline",
      "-Ybackend:GenBCode"),
    //assembly specific overrides
    // jarName in assembly := name.value + "_" + version.value + ".jar",
    jarName in assembly := name.value + ".jar",
    test in assembly := {}
  ).
  settings(scalaxy:_*).
  settings(scalaCheck:_*).
  settings(scalaTest:_*).
  settings(apacheSimplex:_*)

