%% (logic) UFLIA
%% (info) :source 
%% Binary seach algorithm
%% ======================
%% 
%% (declare-datatypes () ((Tree empty (node (val Int) (left Tree) (right Tree)))))
%% 
%% @pre: searchtree(t)
%% @post: res ↔ in(v, t)
%% def binSearch(t: Tree v: Int) =
%%   if (t == empty)
%%      false
%%   else {
%%      if (v == val(t))
%%         true
%%      else if (v < val(t))
%%         binSearch(left(t), v)
%%      else 
%%         binSearch(right(t), v)
%%   }
%% 
%% Partial correctness:
%% 
%% Proof of postcondition by induction, assuming precondition:
%% ∀ t:Tree, v:Int
%%   searchtree(t) →
%%   let res =
%%       if t == empty then
%%          false
%%       else
%%          if v = val(t) then
%%             true
%%          else if v < val(t) then
%%                  let res1 = // res1 is result of recursive call, by I.H.
%%                      let t = left(t)
%%                       in in(v, t) // I.H.
%%                   in res1 // result res is just res1
%%               else
%%                  let res1 = // res1 is result of recursive call, by I.H.
%%                      let t = right(t)
%%                       in in(v, t) // I.H.
%%                   in res1 // result res is just res1
%%    in res ↔ in(v, t)
%% 
%% 
%% Proof precondition of induction hypotheses:
%% ∀ t:Tree, v:Int
%%   searchtree(t) →
%%   if t = empty then
%%      true
%%   else
%%      if v = val(t) then
%%         true
%%      else if v < val(t) then
%%              let t = left(t)
%%               in searchtree(t) 
%%           else
%%              let t = right(t)
%%               in searchtree(t)
%% 

include('swv-bintree.p').

%% Declarations:
tff(searchtree, type, searchtree: 'Tree' > $o).
tff(in, type, in: ($int * 'Tree') > $o).

%% Definition of the in-relation (membership relation) for binary trees
%% ∀ v:Int t:Tree (in(v, t) = (if (t = empty) false else ((v = val(t)) ∨ in(v, left(t)) ∨ in(v, right(t)))))
tff(formula, axiom,
  ( ! [V:$int, T:'Tree'] :
      ( in(V, T)
    <=> ( ( (T = 'empty:Tree')
	 => $false)
	& ( (T != 'empty:Tree')
	 => ( (V = 'val:(Tree)>Int'(T))
	    | in(V, 'left:(Tree)>Tree'(T))
	    | in(V, 'right:(Tree)>Tree'(T)))))))).

%% Definition of binary search trees
%% ∀ t:Tree (searchtree(t) = (if (t = empty) true else (searchtree(left(t)) ∧ searchtree(right(t)) ∧ ∀ v:Int (in(v, left(t)) ⇒ (v ≤ val(t))) ∧ ∀ v:Int (in(v, right(t)) ⇒ (v > val(t))))))
tff(formula, axiom,
  ( ! [T:'Tree'] :
      ( searchtree(T)
    <=> ( ( (T = 'empty:Tree')
	 => $true)
	& ( (T != 'empty:Tree')
	 => ( searchtree('left:(Tree)>Tree'(T))
	    & searchtree('right:(Tree)>Tree'(T))
	    & ( ! [V:$int] :
		  ( in(V, 'left:(Tree)>Tree'(T))
		 => $lesseq(V, 'val:(Tree)>Int'(T))))
	    & ( ! [V:$int] :
		  ( in(V, 'right:(Tree)>Tree'(T))
		 => $greater(V, 'val:(Tree)>Int'(T)))))))))).

%% (Negated) conjecture according to the first proof obligation above
%% ¬∀ t:Tree v:Int (searchtree(t) ⇒ let res=(if (t = empty) false else (if (v = val(t)) true else (if (v < val(t)) let res1=let t=left(t) in in(v, t) in res1 else let res1=let t=right(t) in in(v, t) in res1))) in (res = in(v, t)))
tff(formula, conjecture,
  (  ( ! [T:'Tree', V:$int] :
	 ( searchtree(T)
	=> ( ( ( (T = 'empty:Tree')
	      => $false)
	     & ( (T != 'empty:Tree')
	      => ( ( (V = 'val:(Tree)>Int'(T))
		  => $true)
		 & ( (V != 'val:(Tree)>Int'(T))
		  => ( ( $less(V, 'val:(Tree)>Int'(T))
		      => ( ? [T_3:'Tree'] :
			     ( (T_3 = 'left:(Tree)>Tree'(T))
			     & in(V, T_3))))
		     & ( (~ $less(V, 'val:(Tree)>Int'(T)))
		      => ( ? [T_5:'Tree'] :
			     ( (T_5 = 'right:(Tree)>Tree'(T))
			     & in(V, T_5)))))))))
	 <=> in(V, T)))))).

