% Note that any non-linear term is converted immediately to an NLPP term
% Also, importing these as lemmas makes all variables abstraction.
% This was not the case for the lemmas built into Beagle originally

% Definition with NLPP, as used originally in Beagle:
%% tff(mult_def_1, axiom, ![N: $int]: '#nlpp'(0, N) = 0).
%% tff(mult_def_2, axiom, ![M: $int, N: $int] : '#nlpp'($sum(1, M), N) = $sum(N, '#nlpp'(M, N)) ).

% Definition using product, perhaps simplified as tautologies?
%% tff(mult_def_1, axiom, ![N: $int]: $product(0, N) = 0).
tff(mult_def_2, axiom, ![M: $int, N: $int] : $product($sum(1, M), N) = $sum(N, $product(M, N)) ).

% Basic axioms
tff(mult_comm, axiom, ![A: $int, B: $int]: $product(A,B) = $product(B,A)).
tff(mult_dist, axiom, ![A: $int, B: $int, C: $int]: $product(A, $sum(B,C)) = $sum($product(A, B), $product(A, C))).
% Kills performance
%% tff(mult_assoc, axiom, ![A: $int, B: $int, C: $int]: $product($product(A,B),C) = $product(A, $product(B,C))).

% Lemmas for simplification
%lemma minus_mult_left: "- (a * b) = - a * b"
tff(minus_mult_left, axiom, ![A: $int, B: $int]: $uminus($product(A,B)) = $product($uminus(A),B)).
% conjecture not provable

%% lemma mult_cancel_right1 [simp]: "c = b * c ⟷ c = 0 ∨ b = 1"
tff(mult_cancel_right1, axiom, ![C: $int, B: $int]: ($product(C,B) = C <=> (C = 0 | B = 1))).
% not provable
% provable from comm, mult_eq_0 in 8s

%% lemma mult_pos_pos[simp]: "0 < a ⟹ 0 < b ⟹ 0 < a * b"
tff(mult_nonneg_nonneg, axiom, ![A: $int, B: $int]: (($less(0, A) & $less(0,B)) => $less(0, $product(A,B)))).
% not provable

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Other lemmas, derivable from the above

%% lemma minus_mult_right: "- (a * b) = a * - b"
%% tff(minus_mult_right, conjecture, ![A: $int, B: $int]: $uminus($product(A,B)) = $product(A,$uminus(B))).
% conjecture not provable
% provable from comm and minus_mult_left

%% lemma minus_mult_minus [simp]: "- a * - b = a * b"
%% tff(minus_mult_minus, axiom, ![A: $int, B: $int]: $product($uminus(A),$uminus(B)) = $product(A,B)).
% conjecture not provable
% provable from comm and minus_mult_left

%% lemma right_diff_distrib [algebra_simps]: "a * (b - c) = a * b - a * c"
%% tff(right_diff_distrib, conjecture, ![A: $int, B: $int, C: $int]: $product(A, $difference(B,C)) = $difference($product(A,B),$product(A,C))).
% conjecture not provable
% provable from comm, distr, minus_mult_left (can be proved without, but including it reduces time by about 6s)
%% lemma left_diff_distrib [algebra_simps]: "(a - b) * c = a * c - b * c"
%% skip

%% lemma mult_eq_0_iff [simp]: "a * b = 0 ⟷ a = 0 ∨ b = 0"
%% tff(mult_eq_0_iff, conjecture, ![A: $int, B: $int]: ($product(A,B) = 0 <=> (A = 0 | B = 0))).
% not provable
% provable from comm, mult_cancel_right1 in short time

%% %% lemma mult_cancel_right2 [simp]: "a * c = c ⟷ c = 0 ∨ a = 1"
%% %% tff(mult_cancel_right2, axiom, ![A: $int, C: $int]: ($product(A,C) = C <=> (C = 0 | B = 1))).
% skip

%% lemma mult_cancel_left1 [simp]: "c = c * b ⟷ c = 0 ∨ b = 1"
%% tff(mult_cancel_left1, axiom, ![C: $int, B: $int]: ($product(C,B) = C <=> (C = 0 | B = 1))).

%% skip
%% lemma mult_cancel_left2 [simp]:
%%   "c * a = c ⟷ c = 0 ∨ a = 1"

%% lemma mult_nonneg_nonneg[simp]: "0 ≤ a ⟹ 0 ≤ b ⟹ 0 ≤ a * b"
%% tff(mult_nonneg_nonneg, conjecture, ![A: $int, B: $int]: (($lesseq(0, A) & $lesseq(0,B)) => $lesseq(0, $product(A,B)))).
% not provable
% provable from mult_nonneg_nonneg but not vv

%% lemma zero_le_square [simp]: "0 ≤ a * a"
%% tff(zero_le_square, conjecture, ![A: $int]: $lesseq(0, $product(A,A))).
% provable from mult_nonneg_nonneg

%% lemma not_square_less_zero [simp]: "¬ (a * a < 0)"
%% tff(not_square_less_zero, conjecture, ![A: $int]: ~($less($product(A,A), 0))).
% provable from mult_nonneg_nonneg
